//
// Created by vlad on 03.04.2025.
//

#include "daemon.h"

#include <boost/thread/pthread/thread_data.hpp>


void api_driver::TerminalApiDaemon::connectToApi() {
    {
        std::lock_guard _lock(this->stateMutex);
        this->state.fInitState = "Not connected to API";
    }
#if defined(MODEM_IS_SCPC) || defined(MODEM_IS_TDMA)
    {
        std::lock_guard _lock(this->settingsMutex);
        this->settingsNetwork.loadDefaults();
    }
#endif

    for (int connectAttempt = 0;; connectAttempt++) {
        BOOST_LOG_TRIVIAL(info) << "api_driver::TerminalApiDaemon::connectToApi(): Try to connect api (attempt " << connectAttempt << ")...";
        try {
            cp.connect();

            std::string tmp = cp.getDmaDebug("status_init");
            {
                std::lock_guard _lock(this->stateMutex);
                this->state.fInitState = tmp;
            }
            BOOST_LOG_TRIVIAL(info) << "api_driver::TerminalApiDaemon::connectToApi(): Success connect!";
            BOOST_LOG_TRIVIAL(info) << "api_driver::TerminalApiDaemon::connectToApi(): API status: " << tmp;

            obj::TerminalFirmwareVersion f;
            f.load(cp);

            {
                std::lock_guard _lock(this->firmwareMutex);
                this->firmware = f;
            }

            cp.lastCpError = OK;
            break;
        } catch (std::exception& e) {
            BOOST_LOG_TRIVIAL(error) << "api_driver::TerminalApiDaemon::connectToApi(): connect error " << e.what();
        }
        boost::this_thread::sleep_for(boost::chrono::duration(boost::chrono::milliseconds(1000)));
    }
}

void api_driver::TerminalApiDaemon::run() {
    // это демон, который в бесконечном цикле опрашивает API
    this->connectToApi();

    struct {
        obj::CpUpdatebleObject* u;
        std::function<void ()> c;
    } updaters[] = {
#if API_OBJECT_NETWORK_SETTINGS_ENABLE
        // обновление логов
        {.u = &statsLogs, .c = [this]() {
            try {
                {
                    std::lock_guard _alock(this->cpApiMutex);
                    this->statsLogs.updateCallback(cp);
                }
                BOOST_LOG_TRIVIAL(debug) << "api_driver::TerminalApiDaemon->statsLogs.putItem(): success write statistics state to log!";
            } catch (std::exception& e) {
                BOOST_LOG_TRIVIAL(error) << "api_driver::TerminalApiDaemon->statsLogs.putItem() failed to log: " << e.what();
            }
        }},
#endif
        // обновление статистики
        {.u = state, .c = [this]() {
            try {
                std::lock_guard _alock(this->cpApiMutex);
                std::lock_guard _slock(this->stateMutex);
                state.updateCallback(cp);
                stateDev.updateCallback(cp);
                BOOST_LOG_TRIVIAL(debug) << "api_driver::TerminalApiDaemon::updateState(): success update!";
            } catch (std::exception& e) {
                BOOST_LOG_TRIVIAL(error) << "api_driver::TerminalApiDaemon::updateState(): " << e.what();
            }
        }},
        // обновление кеша настроек
        {.lastUpdate = 0, .periodMs = CACHE_SETTINGS_UPDATE_MS, .c = [this]() {
           try {
               this->updateSettings();
               BOOST_LOG_TRIVIAL(debug) << "api_driver::TerminalApiDaemon::updateSettings(): success update!";
           } catch (std::exception& e) {
               BOOST_LOG_TRIVIAL(error) << "api_driver::TerminalApiDaemon::updateSettings(): " << e.what();
           }
        }},
        // обновление кеша настроек сети (делается отдельно)
        {.lastUpdate = 0, .periodMs = CACHE_SETTINGS_UPDATE_MS, .c = [this]() {
           try {
               this->updateNetworkSettings();
               BOOST_LOG_TRIVIAL(debug) << "api_driver::TerminalApiDaemon::updateNetworkSettings(): success update!";
           } catch (std::exception& e) {
               BOOST_LOG_TRIVIAL(error) << "api_driver::TerminalApiDaemon::updateNetworkSettings(): " << e.what();
           }
        }},
        // обновление кеша QoS
        {.lastUpdate = 0, .periodMs = CACHE_QOS_UPDATE_MS, .c = [this]() {
           try {
               this->updateQos();
               BOOST_LOG_TRIVIAL(debug) << "api_driver::TerminalApiDaemon::updateQos(): success update!";
           } catch (std::exception& e) {
               BOOST_LOG_TRIVIAL(error) << "api_driver::TerminalApiDaemon::updateQos(): " << e.what();
           }
        }}
    };

    while (true) {
        if (this->lastCpError == ERROR || this->lastCpError == TIMEOUT) {
            BOOST_LOG_TRIVIAL(error) << "api_driver::TerminalApiDaemon::run(): close current daemon session caused error " << this->lastCpError;
            CP_Logout(this->sid);
            this->connectToApi();
        }
#ifdef MODEM_IS_SCPC
        updaters[0].periodMs = this->statsLogs.logEn ? this->statsLogs.logPeriodMs.load() : -1;
#endif
        int64_t sleepTime = 60000; // минута по-умолчанию
        auto now = TIME_NOW();
        for (auto& u: updaters) {
            if (u.checkNeedUpdate(now)) {
                auto targetTime = u.lastUpdate + u.periodMs;
                if (targetTime + SLEEP_THRESHOLD <= now && targetTime - SLEEP_THRESHOLD >= now) {
                    u.lastUpdate = targetTime;
                } else {
                    u.lastUpdate = now;
                }

                u.callback();
                now = TIME_NOW();
            }
            if (u.periodMs >= 0) {
                sleepTime = std::min(sleepTime, u.getNextUpdate(now));
            }
        }

        if (sleepTime > 0) {
            boost::this_thread::sleep_for(boost::chrono::duration(boost::chrono::milliseconds(sleepTime)));
        }
    }
}

api_driver::TerminalApiDaemon::TerminalApiDaemon(): daemon([this]() { this->run(); }) {}

void api_driver::TerminalApiDaemon::getState(obj::TerminalState &dest) {
    std::shared_lock _lock(state);
    dest = state;
}

void api_driver::TerminalApiDaemon::getDeviceState(obj::TerminalDeviceState &dest) {
    std::shared_lock _lock(state);
    dest = stateDev;
}

void api_driver::TerminalApiDaemon::getSettingsRxTx(obj::TerminalRxTxSettings &dest) {
}

void api_driver::TerminalApiDaemon::getNetworkSettings(obj::TerminalNetworkSettings &dest) {
}

void api_driver::TerminalApiDaemon::getQosSettings(bool &isEnabled, std::string &json) {
}

void api_driver::TerminalApiDaemon::setSettingsRxTx(obj::TerminalRxTxSettings &s, bool readback) {
}

void api_driver::TerminalApiDaemon::setSettingsDpdi(DPDI_parmeters &s, bool readback) {
}

void api_driver::TerminalApiDaemon::setSettingsBucLnb(buc_lnb_settings &bucLnb, bool readback) {
}

void api_driver::TerminalApiDaemon::setQosSettings(bool enabled, const std::string &str, bool readback) {
}

void api_driver::TerminalApiDaemon::setNetworkSettings(obj::TerminalNetworkSettings &s, bool readback) {
}

api_driver::obj::TerminalFirmwareVersion api_driver::TerminalApiDaemon::getFirmware() {
}

void api_driver::TerminalApiDaemon::resetPacketStatistics() {
}

void api_driver::TerminalApiDaemon::resetDefaultSettings() {
}

api_driver::TerminalApiDaemon::~TerminalApiDaemon() {
    try {
        daemon.interrupt();
        daemon.try_join_for(boost::chrono::seconds(2));
    } catch (std::exception& e) {
        BOOST_LOG_TRIVIAL(error) << "api_driver::~TerminalApiDaemon(): " << e.what();
    }
}




/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 * Этот демон нужен для того, чтобы получать статистику из API, а так же корректно сохранять настройки
 */
class api_driver::TerminalApiDaemon {
    void updateState()  {
        obj::TerminalDeviceState s;

        std::lock_guard lock(this->cpApiMutex);
        try {
            s.updateCallback(cp);
        }

        {
            std::lock_guard lock2(this->stateMutex);
            this->state = s;
#ifdef MODEM_IS_TDMA
            this->state.fInitState = tmpDevState;
#endif
        }
    }

    void updateSettings() {
        modulator_settings mod{};
        // uint32_t modulatorModcod;
        // CP_GetModulatorParams(sid, "modcod", &modulatorModcod);
        demodulator_settings demod{};
#ifdef MODEM_IS_SCPC
        ACM_parameters_serv_ acm{};
#endif
        DPDI_parmeters dpdi{};
        buc_lnb_settings bucLnb{};

        std::lock_guard lock(this->cpApiMutex);
        logCpApiError("api_driver::TerminalApiDaemon::updateSettings()->CP_GetModulatorSettings()", CP_GetModulatorSettings(sid, mod));
        logCpApiError("api_driver::TerminalApiDaemon::updateSettings()->CP_GetDemodulatorSettings()", CP_GetDemodulatorSettings(sid, demod));
#ifdef MODEM_IS_SCPC
        logCpApiError("api_driver::TerminalApiDaemon::updateSettings()->CP_GetAcmParams()", CP_GetAcmParams(sid, &acm));
#endif
        logCpApiError("api_driver::TerminalApiDaemon::updateSettings()->CP_GetDpdiParams()", CP_GetDpdiParams(sid, &dpdi));
        logCpApiError("api_driver::TerminalApiDaemon::updateSettings()->CP_GetBUC_LNB_settings()", CP_GetBUC_LNB_settings(sid, bucLnb));

        {
            std::lock_guard lock2(this->settingsMutex);
            this->modSettings = mod;
            this->demodSettings = demod;
#ifdef MODEM_IS_SCPC
            this->acmSettings = acm;
#endif
            this->dpdiSettings = dpdi;
            this->bucLnbSettings = bucLnb;
        }
    }

    void updateNetworkSettings() {
        obj::TerminalNetworkSettings s;

        std::lock_guard lock(this->cpApiMutex);
        unsafeLoadNetworkSettings(s);

        {
            std::lock_guard lock2(this->networkSettingsMutex);
            this->networkSettings = s;
        }
    }

    void updateQos() {
        bool tmp1; std::string tmp2;
        std::scoped_lock lock{this->cpApiMutex};
        logCpApiError("api_driver::TerminalApiDaemon::updateQos()->CP_GetQoSSettings()", CP_GetQoSSettings(this->sid, tmp2, tmp1));
        {
            std::lock_guard lock2(this->qosSettingsMutex);
            this->qosEnabled = tmp1;
            this->qosClassesJson = tmp2.empty() ? DEFAULT_QOS_CLASSES : tmp2;
        }
    }

public:



    /**
     * Получение статистики, копирует текущие значения в структуры, переданные по указателю. Если передан пустой указатель, копирования не произойдет.
     */
    TerminalDeviceState getState() {
        TerminalDeviceState s;
        {
            std::shared_lock lock(this->stateMutex);
            s = this->state;
        }
        return s;
    }
    /**
     * Получение настроек, копирует текущие значения в структуры, переданные по указателю. Если передан пустой указатель, копирования не произойдет.
     */
    void getSettings(
        modulator_settings* mod, demodulator_settings* demod,
#ifdef MODEM_IS_SCPC
        ACM_parameters_serv_* acm,
#endif
        DPDI_parmeters* dpdi, buc_lnb_settings* bucLnb) {
        if (mod || demod ||
#ifdef MODEM_IS_SCPC
            acm ||
#endif
            dpdi || bucLnb) {
            std::shared_lock lock(this->settingsMutex);
            if (mod) { *mod = this->modSettings; }
            if (demod) { *demod = this->demodSettings; }
#ifdef MODEM_IS_SCPC
            if (acm) { *acm = this->acmSettings; }
#endif
            if (dpdi) { *dpdi = this->dpdiSettings; }
            if (bucLnb) { *bucLnb = this->bucLnbSettings; }
        }
    }

#ifdef MODEM_IS_SCPC
    bool getIsCinC() {
        std::shared_lock lock(this->settingsMutex);
        return modSettings.is_cinc;
    }
#endif

    bool isTest() {
        std::shared_lock lock(this->settingsMutex);
        return modSettings.tx_is_on && (!modSettings.is_carrier
#ifdef MODEM_IS_SCPC
            || this->modSettings.is_test_data
#endif
        );
    }

    void getNetworkSettings(TerminalNetworkSettings& dest) {
        std::shared_lock lock(this->networkSettingsMutex);
        dest = this->networkSettings;
    }

    void getQosSettings(bool& isEnabled, std::string& json) {
        std::shared_lock lock(this->qosSettingsMutex);
        isEnabled = this->qosEnabled;
        json = this->qosClassesJson;
    }

#ifdef MODEM_IS_SCPC
    void setSettingsRxTx(modulator_settings& mod, demodulator_settings& demod, ACM_parameters_serv_& acm, bool readback = true) {
        std::lock_guard lock(this->cpApiMutex);
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_SetDmaDebug(begin_save_config)", CP_SetDmaDebug(sid, "begin_save_config", ""));
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_SetModulatorSettings()", CP_SetModulatorSettings(this->sid, mod));
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_SetDemodulatorSettings()", CP_SetDemodulatorSettings(this->sid, demod));
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_SetAcmParams()", CP_SetAcmParams(this->sid, acm));
        if (readback) {
            logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_GetModulatorSettings()", CP_GetModulatorSettings(this->sid, mod));
            logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_GetDemodulatorSettings()", CP_GetDemodulatorSettings(this->sid, demod));
            logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_GetAcmParams()", CP_GetAcmParams(this->sid, &acm));
            {
                std::lock_guard lock2{this->settingsMutex};
                this->modSettings = mod;
                this->demodSettings = demod;
                this->acmSettings = acm;
            }
        }
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_SetDmaDebug(save_config)", CP_SetDmaDebug(sid, "save_config", ""));
    }
#else
    void setSettingsRxTx(modulator_settings& mod, demodulator_settings& demod, bool readback = true) {
        std::lock_guard lock(this->cpApiMutex);
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_SetDmaDebug(begin_save_config)", CP_SetDmaDebug(sid, "begin_save_config", ""));
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_SetModulatorSettings()", CP_SetModulatorSettings(this->sid, mod));
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_SetDemodulatorSettings()", CP_SetDemodulatorSettings(this->sid, demod));
        if (readback) {
            logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_GetModulatorSettings()", CP_GetModulatorSettings(this->sid, mod));
            logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_GetDemodulatorSettings()", CP_GetDemodulatorSettings(this->sid, demod));
            {
                std::lock_guard lock2{this->settingsMutex};
                this->modSettings = mod;
                this->demodSettings = demod;
            }
        }
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsRxTx()->CP_SetDmaDebug(save_config)", CP_SetDmaDebug(sid, "save_config", ""));
    }
#endif

    void setSettingsDpdi(DPDI_parmeters& s, bool readback = true) {
        std::lock_guard lock(this->cpApiMutex);
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsDpdi()->CP_SetDmaDebug(begin_save_config)", CP_SetDmaDebug(sid, "begin_save_config", ""));
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsDpdi()->CP_SetDpdiParams()", CP_SetDpdiParams(sid, s));
        if (readback) {
            logCpApiError("api_driver::TerminalApiDaemon::setSettingsDpdi()->CP_GetDpdiParams()", CP_GetDpdiParams(this->sid, &s));
            {
                std::lock_guard lock2{this->settingsMutex};
                this->dpdiSettings = s;
            }
        }
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsDpdi()->CP_SetDmaDebug(save_config)", CP_SetDmaDebug(sid, "save_config", ""));
    }

    void setSettingsBucLnb(buc_lnb_settings& bucLnb, bool readback = true) {
        std::lock_guard lock(this->cpApiMutex);
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsBucLnb()->CP_SetDmaDebug(begin_save_config)", CP_SetDmaDebug(sid, "begin_save_config", ""));
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsBucLnb()->CP_SetBUC_LNB_settings()", CP_SetBUC_LNB_settings(this->sid, bucLnb));
        if (readback) {
            logCpApiError("api_driver::TerminalApiDaemon::setSettingsBucLnb()->CP_GetBUC_LNB_settings()", CP_GetBUC_LNB_settings(this->sid, bucLnb));
            {
                std::lock_guard lock2{this->settingsMutex};
                this->bucLnbSettings = bucLnb;
            }
        }
        logCpApiError("api_driver::TerminalApiDaemon::setSettingsBucLnb()->CP_SetDmaDebug(save_config)", CP_SetDmaDebug(sid, "save_config", ""));
    }

    void setQosSettings(bool enabled, const std::string& str, bool readback = true) {
        std::lock_guard lock(this->cpApiMutex);
        logCpApiError("api_driver::TerminalApiDaemon::setQosSettings()->CP_SetDmaDebug(begin_save_config)", CP_SetDmaDebug(sid, "begin_save_config", ""));
        logCpApiError("api_driver::TerminalApiDaemon::setQosSettings()->CP_SetQoSSettings()", CP_SetQoSSettings(this->sid, str, enabled));
        if (readback) {
            bool tmp1; std::string tmp2;
            logCpApiError("api_driver::TerminalApiDaemon::setQosSettings()->CP_GetQoSSettings()", CP_GetQoSSettings(this->sid, tmp2, tmp1));
            {
                std::lock_guard lock2(this->qosSettingsMutex);
                this->qosEnabled = tmp1;
                this->qosClassesJson = tmp2.empty() ? DEFAULT_QOS_CLASSES : tmp2;
            }
        }
        logCpApiError("api_driver::TerminalApiDaemon::setQosSettings()->CP_SetDmaDebug(save_config)", CP_SetDmaDebug(sid, "save_config", ""));
    }

    void setNetworkSettings(TerminalNetworkSettings& s, bool readback = true) {
        // const auto mang = splitIpAndMask();
        // std::pair<std::string, std::string> data;
        // if (!s.isL2) {
        //     data = splitIpAndMask(s.dataIp);
        // }

        std::lock_guard lock(this->cpApiMutex);
        logCpApiError("api_driver::TerminalApiDaemon::setNetworkSettings()->CP_SetDmaDebug(begin_save_config)", CP_SetDmaDebug(sid, "begin_save_config", ""));
        unsafStoreNetworkSettings(s);

        if (readback) {
            unsafeLoadNetworkSettings(s);
            {
                std::lock_guard lock2(this->networkSettingsMutex);
                this->networkSettings = s;
            }
        }
        logCpApiError("api_driver::TerminalApiDaemon::setNetworkSettings()->CP_SetDmaDebug(save_config)", CP_SetDmaDebug(sid, "save_config", ""));
    }

    TerminalFirmwareVersion getFirmware() {
        std::shared_lock lock(this->firmwareMutex);
        return firmware;
    }

    void resetPacketStatistics() {
        std::string tmp;
        std::lock_guard lock(this->cpApiMutex);
        logCpApiError("api_driver::TerminalApiDaemon::resetPacketStatistics()->CP_GetDmaDebug(reset_cnt_rx)", CP_GetDmaDebug(sid, "reset_cnt_rx", &tmp));
    }

    void resetDefaultSettings() {
        std::lock_guard lock(this->cpApiMutex);
        logCpApiError("api_driver::TerminalApiDaemon::resetDefaultSettings()->CP_SetDmaDebug(begin_save_config)", CP_SetDmaDebug(sid, "begin_save_config", ""));
        logCpApiError("api_driver::TerminalApiDaemon::resetDefaultSettings()->CP_SetDmaDebug(default_params)", CP_SetDmaDebug(sid, "default_params", ""));
        logCpApiError("api_driver::TerminalApiDaemon::resetDefaultSettings()->CP_SetDmaDebug(save_config)", CP_SetDmaDebug(sid, "save_config", ""));
    }

};
